import re

from nltk import word_tokenize
from tqdm import tqdm


def clean_text(text, regex_pattern, drop, replace, maxlen=15):
    """ Tokenizes and cleans text, by matching it against regex_pattern and droping and replacing provided patterns. 
    
    Parameters
        text (string): Input text.
        regex_pattern (string): Allowed patterns e.g. [a-zA-Z]
        drop (set): Set of patterns that should be dropeed from text.
        replace (dict): Dictionary of pattern --> replacement.
    
    Returns:
        out_text (string): Tokenized and cleaned up text with respect to all above criteria. 

    """
    if not isinstance(text, str):
        raise TypeError("Input must be a string.")
    if len(text) == 0:
        raise ValueError("Input must be a non empty string.")
    
    # Drop unwanted tokens: Replace them with space, then replace resulting \s{2, } with one space
    for d in drop: # d = e.g. <br/>
        if re.search(d, text):
            text = re.sub(d, ' ', text)
    text = re.sub('\s{2,}', ' ', text)

    for k,v in replace.items():
        if re.search(k, text):
            text = re.sub(k , v, text)

    tokens = word_tokenize(text)
    out_tokens = []
    for t in tokens:
        if len(t) < maxlen:
            if re.match(regex_pattern, t):
                out_tokens.append(t)
    out_text = ' '.join(out_tokens)
    return out_text



class MWE(object):
    """ Multiword Expressions (MWEs) class provides functionalities for an unsupervised extraction of MWEs from text.

    Features 
    TODO: 
    
    1. Functions extract MWEs from training data via PMI, NPMI, and SDMA from training corpus with application of POS tags keeping meaningful sequences (options: keep only NNNNs, JJNN, 
    or keeping all grammatically meaningful)
    2. Function to hyphenate the above MWEs in the corpus before creation of the embeddings so that MWEs get an embedding.
    3. If 2 is not applied, function to create summed embeddings for MWEs extracted in 1, and include them in fasttext/or other we model. 
    4. Function to find above MWEs (extracted in 1) in test data and hyphenate them. 
    """
    
    def __init__(self):
        pass




import os

import pandas as pd

from matplotlib import pyplot as plt
from scipy.stats import zscore
from snlp import logger
from snlp import gaussianize
from sklearn.feature_extraction.text import TfidfVectorizer


class WordFilter(object):
    def __init__(self):
        pass

    def analysis_report(self, documents):
        """Plots word scores/frequency needed for different types of analysis.
    
        """
        vectorizer = TfidfVectorizer(min_df=1)
        X = vectorizer.fit_transform(documents)

    def idf_filterset(self, documents, method='automatic', z=None, l_idf=1, u_idf=8):
        """Creates a filter set by identifying words with anomalous IDF value.

        Args:
            documents (iterable): An iterable which yields either str, unicode or file objects.
            method (string): Method of creating filterset: automatic or manual.
            z (float): Z-score of the IDF distribution above which words are considered anomalous.
            l_idf  (float): Lower cut-off threshold for IDF (inclusive). Used only when 
            u_idf (float): Upper cut-off threshold for IDF (inclusive).

        Returns:
            filter_set (set): Set of filter words
        """

        vectorizer = TfidfVectorizer(min_df=1)
        X = vectorizer.fit_transform(documents)
        
        if method == 'automatic':
            if z:
                filterset = self._create_filter_zscore(vectorizer, zscore_threshold=z)
            else:
                filterset = self._create_filter_zscore(vectorizer)
        elif method == 'manual':
            filterset = self._create_filter(vectorizer, lower_idf=l_idf, upper_idf=u_idf)

        return filterset
        
    def _create_filter(self, vectorizer, lower_idf, upper_idf):
        """

        Args:
            lower_idf  (float): Lower cut-off threshold for IDF (inclusive).
            upper_idf (float): Upper cut-off threshold for IDF (inclusive).

        """

        idf = vectorizer.idf_
        tfidf = dict(zip(vectorizer.get_feature_names(), idf))
        sl = sorted(tfidf.items(), key=lambda kv: kv[1])

        smallest_idf = sl[0][1]
        largest_idf = sl[len(sl)-1][1]

        if lower_idf < smallest_idf:
            raise ValueError("Idf values are between [%.2f, %.2f]. You have set lower_idf to %.2f. Update the values accordingly, \
                or consider continuing with automatic filter creation." 
                %(smallest_idf, largest_idf, lower_idf))
            
        if upper_idf > largest_idf:
            raise ValueError("Idf values are between [%.2f, %.2f]. You have set upper_idf to %.2f. Update the values accordingly, \
                or consider using automatic filter creation." 
                %(smallest_idf, largest_idf, upper_idf))
        
        filter_words = set()
        for i in range(len(sl)):
            if sl[i][1] < lower_idf:
                filter_words.add(sl[i][0])
            if sl[i][1] >= upper_idf:
                filter_words.add(sl[i][0])

        return filter_words


    def _create_filter_zscore(self, vectorizer, zscore_threshold=3):
        idf = vectorizer.idf_
        token_idp_dict = dict(zip(vectorizer.get_feature_names(), idf))
        idf_df = pd.DataFrame(token_idp_dict.items(), columns=['token', 'idf'])

        # Gaussianize idf
        g = gaussianize.Gaussianize(strategy='brute')
        g.fit(idf_df['idf'])
        idf_guassian = g.transform(idf_df['idf'])
        idf_df['idf_gaussianized'] = idf_guassian

        # Calculate z score
        z = zscore(idf_df['idf_gaussianized'])
        idf_df['zscore'] = z

        filter_set = set()
        for i in range(len(idf_df)):
            if idf_df.iloc[i]['zscore'] <= -zscore_threshold or idf_df.iloc[i]['zscore'] >= zscore_threshold:
                filter_set.add(idf_df.iloc[i]['token'])
        return filter_set


    def _create_subplots(self, scores):
        """ Creates subplots corresponding to the histogram of scores. 
        """
        raise NotImplementedError


def filter_text(text, filter_set):
    """
    Args:
        text (list): tokenized text
        filterset (set): Set of filter words
    """

    if not isinstance(text, list):
        raise TypeError("Input must be a list.")
    if len(text) == 0:
        raise ValueError("Input must be a non empty list.")

    res = ' '.join([t for t in text if t not in filter_set])
    return res


def save_filterset_tofile(filter_set, path):
    with open(path, "w") as f:
        for word in filter_set:
            f.write(word + '\n')
    f.close()


def create_filterset_map(p2_raw_train_df, output_dir, zs=[3]):
        original_df = pd.read_csv(p2_raw_train_df, sep='\t', names=['label', 'text'])
        wf = WordFilter()
        z_map = {}
        for z in zs:
            filter_words = wf.idf_filterset(original_df.text, method='automatic', z=z)
            save_filterset_tofile(filter_words, os.path.join(output_dir, 'z'+str(z)+'_filterset.txt'))
            z_map[z] = filter_words
        return z_map


def create_filterset(p2_raw_train_df, output_dir):
        original_df = pd.read_csv(p2_raw_train_df, sep='\t', names=['label', 'text'])
        wf = WordFilter()
        
        filter_words = wf.idf_filterset(original_df.text, method='manual', z=None)
        save_filterset_tofile(filter_words, os.path.join(output_dir, 'manual_filterset.txt'))
        return filter_words



import string
import nltk
import plotly
import pandas as pd
import plotly.figure_factory as ff
import plotly.graph_objs as go

from wordcloud import WordCloud, get_single_color_func
from snlp import logger
from tqdm import tqdm
from nltk.corpus import stopwords
from plotly.subplots import make_subplots


def create_adjust_subplots(labels):
    """Create subplots and adjust the location of the titles wrt the number of labels. 

        Args:
            labels (list): List of (label, type) tuples.

        Returns:
            fig (plotly.graph_objs.Figure)
    """

    if len(labels) == 0:
        titles = ("Analysis of Text", "Analysis of Labels", 
                 "Document Lengths", "", 
                 "", "", "", "Word Frequency", "", "", "", "", ""
                 "Common Nouns", "", "","", "Common Adjectives","", "","", "Common Verbs")

    elif len(labels) == 1:
        titles = ("Analysis of Text", "Analysis of Labels", 
                 "Document Lengths", labels[0][0].capitalize(), "", 
                 "", "Word Frequency", "", "", "", "",
                 "Common Nouns", "", "", "", "Common Adjectives", "", "", "", "Common Verbs")

    elif len(labels) == 2:
        titles = ("Analysis of Text", "Analysis of Labels", 
                 "Document Lengths", labels[0][0].capitalize(), "", labels[1][0].capitalize(),
                 "Word Frequency", "", "", 
                 "Common Nouns", "", "", "", "Common Adjectives", "", "", "", "Common Verbs")
    
    elif len(labels) == 3:
        titles = ("Analysis of Text", "Analysis of Labels", 
                 "Document Lengths", labels[0][0].capitalize(), "", labels[1][0].capitalize(),
                 "Word Frequency", labels[2][0].capitalize(), "", 
                 "Common Nouns", "", "", "", "Common Adjectives", "", "", "", "Common Verbs")

    elif len(labels) == 4:
        titles = ("Analysis of Text", "Analysis of Labels", 
                 "Document Lengths", labels[0][0].capitalize(), "", labels[1][0].capitalize(),
                 "Word Frequency", labels[2][0].capitalize(), "", 
                 "Common Nouns", labels[3][0].capitalize(), "", "Common Adjectives", "", "", "", "Common Verbs")


    fig = make_subplots(rows=16,
                cols=2,
                subplot_titles=titles,
                specs=[[{}, {}],
                    [{"rowspan": 2}, {"rowspan": 2} if len(labels) >= 1 else {}], # row 2
                    [None, None if len(labels) >= 1 else {}], 
                    [{}, {"rowspan": 2} if len(labels) >= 2 else {}],   # row 4
                    [{"rowspan": 2}, None if len(labels) >= 2 else {}],
                    [None, {"rowspan": 2} if len(labels) >= 3 else {}], # row 6
                    [{}, None if len(labels) >= 3 else {}],
                    [{"rowspan": 3}, {"rowspan": 2} if len(labels) >= 4 else {}], # row 8
                    [None, None if len(labels) >= 4 else {}],
                    [None, {}],
                    [{"rowspan": 3}, {}], # 12
                    [None, {}],
                    [None, {}],
                    [{"rowspan": 3}, {}], # 16
                    [None, {}],
                    [None, {}]
                    ],
                vertical_spacing=0.035)
    return fig


def plotly_wordcloud(token_count_dic):
    """Create a world cloud trace for plotly. 

        Args:
            token_count_dic (dictionary): Dictionary of token to its count

        Returns:
            trace (Plotly trace)
    """

    wc = WordCloud(color_func=get_single_color_func('deepskyblue'), max_words=100)
    wc.generate_from_frequencies(token_count_dic) 
    
    word_list=[]
    rel_freq_list=[]
    freq_list = []
    fontsize_list=[]
    position_list=[]
    orientation_list=[]
    color_list=[]


    for (word, rel_freq), fontsize, position, orientation, color in wc.layout_:
        word_list.append(word)
        rel_freq_list.append(rel_freq)
        freq_list.append(token_count_dic[word])
        fontsize_list.append(fontsize)
        position_list.append(position)
        orientation_list.append(orientation)
        color_list.append(color)
        
    # get the positions
    x=[]
    y=[]
    for i in position_list:
        x.append(i[0])
        y.append(i[1])
            
    # get the relative occurence frequencies
    new_freq_list = []
    for i in rel_freq_list:
        new_freq_list.append(i*100)
    new_freq_list
    
    trace = go.Scatter(x=x, 
                       y=y, 
                       textfont = dict(size=new_freq_list,
                                       color=color_list),
                       hoverinfo='text',
                       hovertext=['{0}: {1}'.format(w, f) for w, f in zip(word_list, freq_list)],
                       mode='text',  
                       text=word_list
                      )
    return trace


def generate_text_plots(figure, doc_length_list, word_freq_list, noun_freq_dict, adj_freq_dic, verb_freq_dic):
    """Generate distribution plots and word clouds for the textual content based on the input argumnets.
    
    Args:
        figure (plotly.graph_objs.Figure): Figure object in which the plots are created. 
        doc_length_list (list): List containing the length of each document. 
        word_freq_list (list): List containing the frequency of each word. 
        noun_freq_dict (dictionary): Dictionary of noun to frequency.
        adj_freq_dict (dictionary): Dictionary of noun to frequency.
        verb_freq_dict (dictionary): Dictionary of verb to frequency.

    Returns:
        None

    """

    # Customize plots and x, y labels: https://plotly.com/python/subplots/#customizing-subplot-axes
    # Word cloud with plotly: https://github.com/PrashantSaikia/Wordcloud-in-Plotly/blob/master/plotly_wordcloud.py
    
    def _distplot_to_dist_trace(data, color):
        data_dist = ff.create_distplot([data], group_labels=['distplot'], colors = [color])['data']
        for item in data_dist:
            item.pop('xaxis', None)
            item.pop('yaxis', None)
        return data_dist 
    
    data1_dist = _distplot_to_dist_trace(doc_length_list, color='rgb(0, 200, 200)')
    data2_dist = _distplot_to_dist_trace(word_freq_list, color='magenta')
            
    d1_hist = data1_dist[0]
    d1_kde = data1_dist[1]
    d1_rug = data1_dist[2]

    d2_hist = data2_dist[0]
    d2_kde = data2_dist[1]
    d2_rug = data2_dist[2]

    noun_cloud = plotly_wordcloud(token_count_dic=noun_freq_dict)
    adj_cloud = plotly_wordcloud(token_count_dic=adj_freq_dic)
    verb_cloud = plotly_wordcloud(token_count_dic=verb_freq_dic)

    figure.append_trace(d1_hist, 2, 1)
    figure.append_trace(d1_kde, 2, 1)
    figure.append_trace(d1_rug, 4, 1)

    figure.append_trace(d2_hist, 5, 1)
    figure.append_trace(d2_kde, 5, 1)
    figure.append_trace(d2_rug, 7, 1)
    
    figure.append_trace(noun_cloud, 8, 1)
    figure.append_trace(adj_cloud, 11, 1)
    figure.append_trace(verb_cloud, 14, 1)

    figure.update_xaxes(rangemode='tozero', row=4, col=1)
    figure.update_yaxes(showticklabels=False, row=4, col=1)

    figure.update_xaxes(rangemode='tozero', row=7, col=1)
    figure.update_yaxes(showticklabels=False, row=7, col=1)

    figure.update_yaxes(title_text="Probability", row=2, col=1)
    figure.update_yaxes(title_text="Probability", row=5, col=1)
    
    figure.update_xaxes(showticklabels=False, zeroline=False, row=8, col=1)
    figure.update_xaxes(showticklabels=False, zeroline=False, row=11, col=1)
    figure.update_xaxes(showticklabels=False, zeroline=False, row=14, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=8, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=11, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=14, col=1)


def generate_label_plots(figure, df, label_cols):
    """Generate histogram and bar plots for the labels in label_cols.
    
    Args:
        figure (plotly.graph_objs.Figure): Figure object in which the plots are created. 
        df (Pandas DataFrame): DataFrame that contains labels specified in label_cols.
        label_cols (list): list of tuples in the form of [('label_1', 'categorical/numerical'), ('label_2', 'categorical/numerical'), ...]
        
    Returns:
        None

    """
    if len(label_cols) == 1:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
    elif len(label_cols) == 2:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
    elif len(label_cols) == 3:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        lab_trace3 = label_plot(df, label_col=label_cols[2][0], label_type=label_cols[2][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.append_trace(lab_trace3, 6, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
        figure.update_yaxes(title_text="Count", row=6, col=2)
    elif len(label_cols) == 4:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        lab_trace3 = label_plot(df, label_col=label_cols[2][0], label_type=label_cols[2][1])
        lab_trace4 = label_plot(df, label_col=label_cols[3][0], label_type=label_cols[3][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.append_trace(lab_trace3, 6, 2)
        figure.append_trace(lab_trace4, 8, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
        figure.update_yaxes(title_text="Count", row=6, col=2)
        figure.update_yaxes(title_text="Count", row=8, col=2)


def label_plot(df, label_col, label_type):
    """Creates a plot for label_col in df, wrt to label_type. 

    Args:
        df (Pandas DataFrame): DataFrame that contains label_col. 
        label_col (string): Name of the label column in df that must be plotted.
        label_type (string): Represents the type of label and consequently specifies the type of plot. It can be "numerical" or "categorical". 
    
    Returns:
        trace (Plotly trace)

    """

    if label_type == 'categorical':
        values = df[label_col].unique().tolist() # ['pos', 'neg', 'neutral']
        counts = df[label_col].value_counts() # 1212323
        x = []
        y = []
        for v in values:
            x.append(v)
            y.append(counts[v])
        trace = go.Bar(x=x, y=y, name=label_col)
    elif label_type == 'numerical':
        trace = go.Histogram(x=df[label_col], name=label_col)
    else:
        raise ValueError('label_col input argument must be set to either \"categorical\" or \"numerical\".')
    return trace


def generate_report(df, out_dir, text_col, label_cols=[], language='english', skip_stopwords_punc=True, save_report=False):
    """Generate analysis report and eitherr renders the report via Plotly show api or saves it offline to html. 

    Args:
        df (Pandas DataFrame): DataFrame that contains text and labels. 
        text_col (string): Name of the column that contains a tokenized text content.
        label_cols (list): list of tuples in the form of [('label_1', 'categorical/numerical'), ('label_2', 'categorical/numerical'), ...]
        language (string): Language of the text in df[text_col]
        skip_stopwords_punc (bool): Whether or not skip stopwords and punctuations in the analysis. Default: True
        save_report (bool): Whether or not save the report as an html file. Default: False
    
    Returns:
        None

    """

    def update_count(items_dic, items):
        for t in items:
            if t in items_dic:
                items_dic[t] += 1
            else:
                items_dic[t] = 1

    def get_pos(tagged_tokens, pos):
        res = []
        for pt in tagged_tokens:
            if pt[1].startswith(pos):
                res.append(pt[0])
        return res

    if len(label_cols)>4:
        raise ValueError('Maximum of 4 labels can be specidied for analysis.')

    stop_words = set(stopwords.words(language))
    punctuations = set(string.punctuation)

    doc_lengths = []
    token_to_count = {}
    NNs = {}
    JJs = {}
    Vs = {}    
    
    logger.info('Processing text in %s column of the input DataFrame...' %text_col)
    for text in tqdm(df[text_col]):
        try:
            tokens = text.lower().split(' ')
            doc_lengths.append(len(tokens))
            if skip_stopwords_punc:
                tokens = [t for t in tokens if t not in stop_words and t not in punctuations]
                update_count(token_to_count, tokens)

        except Exception as e:
            logger.warning('Processing entry --- %s --- lead to exception: %s' %(text, e.args[0]))
            continue
        
        postag_tokens = nltk.pos_tag(tokens)
        nouns = get_pos(postag_tokens, 'NN')
        update_count(NNs, nouns)
        verbs = get_pos(postag_tokens, 'VB')
        update_count(Vs, verbs)
        adjectives = get_pos(postag_tokens, 'JJ')
        update_count(JJs, adjectives)

    word_frequencies = [v for _, v in token_to_count.items()]

    fig_main = create_adjust_subplots(label_cols)

    logger.info('Generating distplots and word cloud for input text')
    generate_text_plots(fig_main, doc_lengths, word_frequencies, NNs, JJs, Vs)
    logger.info('Generating plots for labels')
    generate_label_plots(fig_main, df, label_cols)
    logger.info('Rendering plots')
    fig_main.update_layout(height=3100, showlegend=False)
    if save_report:
        plotly.offline.plot(fig_main, filename='report.html')
    else:
        fig_main.show()



import json
import collections
import pandas as pd
import re
import tqdm
import time


def replace_compunds(path_to_compounds, df, text_column, am_threshold=0.5, only_compounds=False, lower_case=False):
    """Hyphenates the compounds in the corpus so that they are treated as a single token by downstream applications.

    Args:
        
        path_to_nns (string): Path to the ranked list of compounds
        df (pd.FataFrame): DataFrame comprising training data with a text column
        text_column (string): Text (content) column of df
        am_threshold (float): AM threshold above which the compounds are selected for replacement
        only_compounds (bool): Whether or not keep only compounds and drop the rest of the text

    Returns:
        df (pd.FataFrame)

    """

    json_file = open(path_to_compounds)
    json_str = json_file.read()

    pmi_ord_dict = json.loads(json_str, object_pairs_hook=collections.OrderedDict)
    print('Num of all compounds: %d' %len(pmi_ord_dict))
    good_nns = set()
    
    # Current format of pmi_ord_dict: [['victor jones', [14.72, 0.96]], ...]
    # pmi_ord_dict[i][1][1]: npmi  | pmi_ord_dict[i][1][0]: pmi   |  pmi_ord_dict[i][0]: compound
    i = 0
    while  float(pmi_ord_dict[i][1][1]) > am_threshold:
        good_nns.add(pmi_ord_dict[i][0])
        i += 1
    
    print('Number of compounds to be replaced in corpus: %d' %len(good_nns))

    new_text = []
    for sent in tqdm.tqdm(df[text_column]):
        # Extract bigrams from text
        sent = (sent.lower() if lower_case else sent)
        bigrams = get_ngrams(sent, 2)
        
        if only_compounds:
            tmp = ''
            for bg in bigrams:
                if bg in good_nns:
                    tmp += (bg.split(' ')[0]+'-'+bg.split(' ')[1]+' ')
            if tmp != '':
                sent = tmp.strip()
        else:
            for bg in bigrams:
                if bg in good_nns:
                    sent = re.sub(bg, bg.split(' ')[0]+'-'+bg.split(' ')[1], sent)
        new_text.append(sent)
    df[text_column] = new_text
    return df


def get_ngrams(sentence, n):
    """Extracts n-grams from sentence.
    
    Args:
        sentence (string)
        n (int)
        lower(bool)

    Returns:
        ngrams(list)
    """
    #TODO handle excecases when sent is nan, here and in the calling function
    tokens = sentence.split(" ")
    ngrams = []
    for i in range(len(tokens)-n+1):
        ngrams.append(' '.join(tokens[i:i+n]))
    return ngrams


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="snlp", # Replace with your own username
    version="0.0.6",
    author="meghdadFar",
    author_email="meghdad.farahmand@gmail.com",
    description="Statistical NLP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/meghdadFar/snlp",
    packages=setuptools.find_packages(),
    # install_requires = ['pandas==1.0.3', 'scikit-learn==0.22.2.post1', 
    # 'matplotlib==3.2.1', 'scipy==1.4.1', 'plotly==4.8.1',
    # 'nltk==3.5', 'tqdm==4.45.0'],
    scripts=['bin/downloads.py'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import nltk
import tqdm
import json
import math
import re
import os

import pandas as pd

from nltk.tokenize import word_tokenize
from nltk.tokenize import ToktokTokenizer
from collections import Counter
from snlp import logger 


"""
TODO: 
    
    1. Include more AMs
    2. Include SDMA
    3. Function to create embedding for MWEs by combining the embeding of their components. 
"""


def extract_ncs_from_sent(sentence):
    """Extract two-word noun compounds from tokenized input. 
    
    Args:
        sentence (string): A tokenized sentence. 

    Returns:
        NNS_counts (dict): Dictionary of noun-noun compounds to their count in sentence. 
        JNs_counts (dict): Dictionary of adjective-noun compounds to their count in sentence. 
        word_counts (dict) Dictionary of words to their count in sentence. 
    """
    if not isinstance(sentence, str):
        raise TypeError('Input must be a string.')
    
    if len(sentence) == 0:
        return
        
    tokens = sentence.split(" ")
    postag_tokens = nltk.pos_tag(tokens)
    JNs = []
    NNs = []
    for i in range(len(postag_tokens)-1):
        w1 = postag_tokens[i]
        if w1[1] not in ['NN', 'NNS', 'JJ']:
            continue
        else:
            w2 = postag_tokens[i+1]
            if not re.match("[a-zA-Z0-9]{2,}", w1[0]) or not re.match("[a-zA-Z0-9]{2,}", w2[0]):
                continue

            if w2[1] in ['NN', 'NNS']:
                if i+2 < len(postag_tokens):
                    w3 = postag_tokens[i+2]
                    if w3 not in ['NN', 'NNS']:
                        if w1[1] == 'NN':
                            NNs.append(w1[0]+' '+w2[0])
                        if w1[1] == 'JJ':
                            JNs.append(w1[0]+' '+w2[0])
                else:
                    if w1[1] == 'NN':
                        NNs.append(w1[0]+' '+w2[0])
                    if w1[1] == 'JJ':
                        JNs.append(w1[0]+' '+w2[0])
    
    NNS_counts = Counter(NNs) 
    JNs_counts = Counter(JNs) 
    word_counts = Counter(tokens)
    return NNS_counts, JNs_counts, word_counts


def calculate_pmi(nn_dict, word_dic, N_compounds, N_words):
    """Calculates Pointwise Mutual Information between the two words of every word pair in nn_dict. 

    Args:
        nn_dict (dict): Dictionary of compounds and their count. 
        word_dic (dict): Dictionary of words and their count. 
        N_compounds (int): Number of compounds.
        N_words (int): Number of words. 
    
    Returns:
        nn_dict (dict): Dictionary of compounds and their pmi and npmi values, sorted based on their npmi. 

    """
    pmi = 0.0
    npmi = 0.0
    for compound, count in nn_dict.items():
        w1w2 = compound.split(" ")
        # To filter out compounds that are rare/unique because of strange/misspelled component words. 
        if float(word_dic[w1w2[0]]) > 10 and float(word_dic[w1w2[1]]) > 10:
            p_of_c = float(count) / float(N_words)
            p_of_h = float(word_dic[w1w2[0]]) / float(N_words)
            p_of_m = float(word_dic[w1w2[1]]) / float(N_words)

            pmi = math.log(p_of_c/(p_of_h * p_of_m))
            npmi = pmi/float(-math.log(p_of_c))
            nn_dict[compound] = [round(pmi, 2), round(npmi, 2)]
        else:
            nn_dict[compound] = [0.0, 0.0]

    nn_dict = sorted(nn_dict.items(), key=lambda e: e[1][1], reverse=True)
    return nn_dict


def get_counts(df, text_column, output_dir):
    """Reads a corpus in df (pd.DataFrame) and generates all counts necessary for calculating AMs.

    Args:
        output_dir (string): Path to output dir where calculated counts will be saved. 

    Returns:
        0 (int): If count files are successfully stored in the specified path. 
    """
    NNs_total = {}
    JNs_total = {}
    W_total = {}

    NN_count = 0
    JN_count = 0
    W_count = 0

    for sent in tqdm.tqdm(df[text_column]):
        NNs, JNs, Ws = extract_ncs_from_sent(sent)
        
        for k,v in NNs.items():
            NN_count += v
            if k in NNs_total:
                NNs_total[k] += v
            else:
                NNs_total[k] = v
        
        for k, v in JNs.items():
            JN_count += v
            if k in JNs_total:
                JNs_total[k] += v
            else:
                JNs_total[k] = v
        
        for k, v in Ws.items():
            W_count += v
            if k in W_total:
                W_total[k] += v
            else:
                W_total[k] = v

    try:
        with open(os.path.join(output_dir, 'nns.json'), 'w') as file:
            file.write(json.dumps(NNs_total))

        with open(os.path.join(output_dir, 'jns.json'), 'w') as file:
            file.write(json.dumps(JNs_total))

        with open(os.path.join(output_dir, 'words.json'), 'w') as file:
            file.write(json.dumps(W_total))

        with open(os.path.join(output_dir, 'jn_count.txt') , 'w') as file:
            file.write(str(JN_count))
        
        with open(os.path.join(output_dir, 'nn_count.txt'), 'w') as file:
            file.write(str(NN_count))

        with open(os.path.join(output_dir, 'w_count.txt'), 'w') as file:
            file.write(str(W_count))
    except:
        raise
    else:
        return 0



def get_ams(path_to_counts, ams=['pmi']):
        """Reads the counts from path_to_counts and accordingly, calculates measures specified in ams.
        
        Args:
            path_to_counts (string): Path to counts dir, generated by get_counts. 
            ams (list): List of AMs to be calcuated. 
        
        Returns:
        
        """
        f = open(os.path.join(path_to_counts, 'w_count.txt'))
        v = f.read()
        w_count = int(v)

        f = open(os.path.join(path_to_counts, 'nn_count.txt'))
        v = f.read()
        nn_count = int(v)

        f = open(os.path.join(path_to_counts, 'jn_count.txt'))
        v = f.read()
        jn_count = int(v)

        json_file = open(os.path.join(path_to_counts, 'jns.json'))
        json_str = json_file.read()
        jns = json.loads(json_str)

        json_file = open(os.path.join(path_to_counts, 'nns.json'))
        json_str = json_file.read()
        nns = json.loads(json_str)

        json_file = open(os.path.join(path_to_counts, 'words.json'))
        json_str = json_file.read()
        words = json.loads(json_str)

        if 'pmi' in ams:
            res = calculate_pmi(nn_dict=nns, word_dic=words, N_compounds=nn_count, N_words=w_count)
            with open(os.path.join(path_to_counts, 'nn_pmi.json'), 'w') as file:
                file.write(json.dumps(res))

            res = calculate_pmi(nn_dict=jns, word_dic=words, N_compounds=jn_count, N_words=w_count)
            with open(os.path.join(path_to_counts, 'jn_pmi.json'), 'w') as file:
                file.write(json.dumps(res))
        
        if 'chis' in ams:
            print('Chis is not implemented.')
        
        if 't' in ams:
            print('t is not implemented.')


    